% Matlab code example for Algorithm 4.3 presented in
% "A deterministic sparse FFT algorithm for vectors with small support"
% by Gerlind Plonka and Katrin Wannenwetsch

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Fast implementation of the inverse Fourier Transform of vectors with
% short support where the Fourier data might be noisy.
%
% xrec = algorithm(data,m)
%
% This function returns the inverse Fourier Transform "xrec" for given
% Fourier data "data" with noise. The reconstruction is assumed to have
% short support where "m" is an upper bound.

function xrec=algorithm(data,m)

%compute N, L and J
N=length(data);
J=log2(N);
L=ceil(log2(m));

if L==J || L==J-1
    %inverse FFT
    disp('Support to large! Using standard ifft.')
    xrec=ifft(data);
    
else
    
    % calculate absolute value of Fourier data (for faster maximum
    % calculation later on)
    absdata = abs(data);
    
    %compute x^{(L+1)} by inverse FFT
    xa=ifft(data((2^(J-L-1)*(0:2^(L+1)-1))+1));
   
    %compute energies in xa
    b=zeros(2^(L+1),1);
    b(1)=sum(abs(xa(1:m)).^2);
    for k=2:2^(L+1)
       b(k)=b(k-1)+abs(xa(mod(m+k-2,2^(L+1))+1))^2-abs(xa(k-1))^2;
    end
    
    muvector=zeros(2,1);
    muvector(1)=find(b==max(b),1)-1;      %compute argmax e_k
    
    
    %"shifted" Fourier data
    yL2=zeros(2^(L+1),1);
    yL2(1:2^(L+1))=data((2^(J-L-2)*(2*(1:2^(L+1))-1))+1);
    
    z2=ifft(yL2);
    
    %matrix with xa, z2 as columns
    zvectors=[xa,z2];
    
    %compute energies in z2
    c=zeros(2^(L+1),1);
    c(1)=sum(abs(z2(1:m)).^2);
    for k=2:2^(L+1)
       c(k)=c(k-1)+abs(z2(mod(m+k-2,2^(L+1))+1))^2-abs(z2(k-1))^2;
    end
    
    c=b+c;
    
    %compute \mu_1^{(L+1)}
    muvector(2)=find(c==max(c),1)-1;
    
    t=1;
    while muvector(t) ~= muvector(t+1)

        yLk=zeros(2^(L+1),1);
        yLk(1:2^(L+1))=data(mod(2^(J-L-1)*(0:2^(L+1)-1)+t,N)+1);         

        zk=ifft(yLk);
        
        zvectors(:,t+2)=zk;

        d=zeros(2^(L+1),1);
        d(1)=sum(abs(zk(1:m)).^2);
        for k=2:2^(L+1)
           d(k)=d(k-1)+abs(zk(mod(m+k-2,2^(L+1))+1))^2-abs(zk(k-1))^2;
        end

        c=c+d;
        muvector(t+2)=find(c==max(c),1)-1;
        
        t=t+1;
    end
    
    mu=muvector(end);
    muL=mu;
    
    
    %vector xs is the support of x
    xs=xa(mod(mu+(0:m-1),2^(L+1))+1);       

    for j=L+1:J-1
        
        %find maximal (odd-indexed) component of \hat{x}^(j+1)}
        ind = (2^(J-j-1)+1):(2^(J-j)):(2^J-2^(J-j-1)+1);
        [~,s]=max(absdata(ind));
        xhatmax=data(ind(s));

        %compute a
        a=exp(-pi*1i*(2*s-1)*(mu+(0:m-1))/2^j)*xs;

        if abs(a-xhatmax)>abs(a+xhatmax)     %if Fourier value is closer to -a, then shift mu
            mu=mu+2^j;
        end

    end
    

    
    %improve values of support vector using additional vectors from step 2
    zvectors=[zvectors;zvectors];
    zvectors=zvectors((muL+1):(muL+m),:);
    
    omegamatrix=zeros(m,t+1);
    omegamatrix(:,1)=ones;
    omegamatrix(:,2)=exp(2*pi*1i/2^(L+2)*(mu+(0:m-1)));
    for k=3:t+1
       omegamatrix(:,k)=exp(2*pi*1i/N*(k-2)*(mu+(0:m-1))); 
    end
    
    zvectors=zvectors.*omegamatrix;
    xs=sum(zvectors,2)/(t+1);         %improved support vector
    
 
    

    %build reconstruktion xres of x
    xrec=zeros(N,1);
    xrec(mod(mu+(0:m-1),2^J)+1)=xs(1:m);
    
end

end